package test.springbootsigterm;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.SpringApplication;
import org.springframework.boot.autoconfigure.SpringBootApplication;
import org.springframework.boot.builder.SpringApplicationBuilder;
import org.springframework.boot.context.event.ApplicationEnvironmentPreparedEvent;
import org.springframework.context.event.GenericApplicationListener;
import org.springframework.core.env.PropertySource;
import org.springframework.core.env.StandardEnvironment;

import java.io.IOException;
import java.util.concurrent.atomic.AtomicBoolean;

@SpringBootApplication
public class SpringBootSigtermApplication {

    public static void main(String[] args) {
        // Use StandardEnvironment, otherwise ApplicationEnvironment will be created,
        // and it wraps property sources into SpringConfigurationPropertySource, which
        // swallows exception and doesn't allow modification of property source after
        // environment was created.
        new SpringApplicationBuilder(SpringBootSigtermApplication.class)
            .environment(new StandardEnvironment())
            .listeners(GenericApplicationListener.forEventType(ApplicationEnvironmentPreparedEvent.class, (event) -> {
                // All properties, required by log4j will be read here, before blocking
                // property source is registered.
                Logger logger = LoggerFactory.getLogger(SpringBootSigtermApplication.class);
                logger.info("{}", event);

                // New property source is registered, it will throw an exception
                TestPropertySource propertySource = new TestPropertySource("test");
                event.getEnvironment().getPropertySources().addLast(new TestPropertySource("test"));
                SpringApplication.getShutdownHandlers().add(propertySource::close);

                new Thread(() -> {
                    try {
                        Thread.sleep(100);
                    } catch (InterruptedException e) {
                        Thread.currentThread().interrupt();
                    }
                    long pid = ProcessHandle.current().pid();
                    logger.info("Sending sigterm to myself: {}", pid);
                    try {
                        Runtime.getRuntime().exec(String.format("kill -SIGTERM %s", pid));
                    } catch (IOException e) {
                        throw new RuntimeException(e);
                    }
                }).start();

                try {
                    Thread.sleep(100_000);
                } catch (InterruptedException e) {
                    Thread.currentThread().interrupt();
                }
            }))
            .run(args);
    }

    static class TestPropertySource extends PropertySource<Object> implements AutoCloseable {

        final AtomicBoolean isClosed = new AtomicBoolean(false);

        public TestPropertySource(String name) {
            super(name);
        }

        @Override
        public String getProperty(String name) {
            if (isClosed.get()) {
                System.out.println("Uh, oh, I'm closed, take your null and go away.");
                return null;
            } else {
                throw new RuntimeException("Dear property '" + name + "', can't you see, I'm busy? Stand in line and wait.");
            }
        }

        @Override
        public void close() {
            isClosed.set(true);
        }
    }

}
